import gmpy

from roots.nt import newtons_method as root_of


class PolyHash:

    # Polynomial to compute root from
    #POLYNOMIAL  = r'1x^5+1x^4+%(BLOCK)sx^3+1x^2+1x^1+1x^0'
    PRECISION   = 20

    # Number of bytes to input
    BLOCK_SIZE  = 64

    # Number of bytes to output
    DIGEST_SIZE = BLOCK_SIZE/2

    # Initial value of state?
    INIT        = 'POLYHASHv0.1-256'.ljust(DIGEST_SIZE, '\0')

    # How many bytes are used to store length?
    LENGTH_SIZE = 8


    # test vectors
    VECTORS_PREC_18 = [
        ('', 
         '319de7fa1d26b8674cc8f0f635afb5fa55c64d89bbc1a92cea546c1f4872e864'),
        ('abc',
         'de9fafbd2a29bc5dcea273665af3986525927c3e93a6712bf3f31a9be87c381e'),
        ('cba', 
         '97e67db4eb1bda5bbcd39f021a61628462114af534f4eb97e234376e33f50dbf'),
        ('abcdef',
         '43898a21beef1ea2f830c25b73ed8d5cec86fc612a7daf5a5932035961b203c8'),
        ('The quick brown fox jumps over the lazy dog',
         '450c5c368bd30edd29b7dbe4ee316dcc66da2650a1d52aa8dbe83eaf514f1293'),
        ('The quick crown fox jumps over the lazy dog',
         'cefdbeca2026892628cfae6231c08edda09da50aa6acf91831c42f728b28628b'),
        ('The quick brown fox jumps over the lazy dog'*3,
         '8337714562e63ec9e4c0d3edaa0e40386b8dbf0abeda1ba47fe5ebd6ae03b987'),
        ]


    @staticmethod
    def test_p18():
        global PRECISION
        _p = PRECISION
        PRECISION=18
        for txt, chash in PolyHash.VECTORS_PREC_18:
            print 'vector[%3i]: "%s"' % (len(txt),txt)
            hash0 = PolyHash(txt).hexdigest()
            p = PolyHash()
            for i in xrange(0, len(txt), 7):
                p.update(txt[i:i+7])
            hash1 = p.hexdigest()
            print 'result[%3i]: "%s"' % (len(hash0), hash0)
            assert hash0 == chash
            assert hash1 == chash
        PRECISION = _p


    @staticmethod
    def hex_to_bin(some_hex):
        return ''.join((chr(eval('0x'+some_hex[i:i+2])) for i in xrange(0, len(some_hex), 2)))

    @staticmethod
    def bin_to_hex(some_bin):
        return ''.join((hex(ord(c))[2:].rjust(2,'0')) for c in some_bin)

    @staticmethod
    def block_to_long(block):
        return gmpy.mpz(''.join(map(lambda x: hex(ord(x))[2:].rjust(2,'0'), list(block))),16).setbit(32)

    @staticmethod
    def hash_block(block):
        assert len(block) <= PolyHash.BLOCK_SIZE

        DS = PolyHash.DIGEST_SIZE*8

        # compute polynomial
        block = block.rjust(PolyHash.BLOCK_SIZE, '0')
        coefs = tuple((PolyHash.block_to_long(block[i:i+4])) for i in xrange(0, len(block), 4))

        # substitute coefficient in polynomial
        #poly = ('1x^%i+' % (len(coefs)+1)) \
        #    +  ('+'.join('%ix^%i' % (coef,i+1) for i,coef in enumerate(coefs))) \
        #    +  ('+1x^0')

        poly = ('1x^%i+' % (len(coefs)*2+2)) \
            +  ('+'.join('%ix^%i' % (coef,i*2+2) for i,coef in enumerate(coefs))) \
            +  ('-1x^0')
        print poly

        # compute root
        M = 2**PolyHash.PRECISION
        N, root = root_of(poly, PolyHash.PRECISION, init=1


, use_mp=False)
        digits = root.digits(2)[:2**PolyHash.PRECISION]
        #print digits[:50]
        #print digits[-10:]
        #digits = digits[-PolyHash.DIGEST_SIZE*8:]
        #N, root = root_of(poly, PolyHash.PRECISION+1, init=0.000001, use_mp=False)
        #digits = root.digits(2)[:2**PolyHash.PRECISION]
        #print digits[:50]
        #print digits[-10:]
        #import sys; sys.exit()
        #digits = digits[digits.index('.'):digits.rindex('e')][:N]
        
        # return hexdecimal hash
        return PolyHash.hex_to_bin(gmpy.mpz(digits[-DS:], 2).digits(16)[2:])



    def __init__(self, buffer=''):
        self._state  = PolyHash.INIT
        self._length = 0
        self._buffer = ''
        self.update(buffer)


    def _encode_length(self):
        length = self._length + len(self._buffer)
        hexlen = gmpy.mpz(length).digits(16)[2:].rjust(PolyHash.LENGTH_SIZE*2, '0')
        assert len(hexlen) == PolyHash.LENGTH_SIZE*2
        return PolyHash.hex_to_bin(hexlen)

    
    def _update_block(self, block):
        assert len(block) == PolyHash.DIGEST_SIZE
        self._length += len(block)
        self._state   = PolyHash.hash_block(self._state + block)


    def update(self, msg=''):
        # cowardly refuse to update when less than one block
        if len(self._buffer)+len(msg) <= PolyHash.DIGEST_SIZE:
            self._buffer += msg
            return

        msg = self._buffer + msg
        self._buffer = ''
        for i in xrange(0, len(msg)-PolyHash.DIGEST_SIZE+1, PolyHash.DIGEST_SIZE):
           self._update_block(msg[i:i+PolyHash.DIGEST_SIZE]) 
        self._buffer = msg[i+PolyHash.DIGEST_SIZE:]


    def digest(self):
        # copy values to avoid changing state
        suffix = self._encode_length()
        state  = self._state
        buffer = self._buffer

        # adjust buffer
        if len(buffer) + len(suffix) + 1 > PolyHash.DIGEST_SIZE:
            state  = PolyHash.hash_block(state + buffer.ljust(PolyHash.DIGEST_SIZE, '\0'))
            suffix = '\1'+suffix.rjust(PolyHash.DIGEST_SIZE-1, '\0')
            buffer = ''
        else:
            buffer = buffer.ljust(PolyHash.DIGEST_SIZE-len(suffix)-1, '\0') + '\1'

        return PolyHash.hash_block(state + buffer + suffix)
        
    
    def hexdigest(self):
        return PolyHash.bin_to_hex(self.digest())
