import sys

from math import floor,ceil,log
log2 = (lambda x: log(x) / log(2))

from database import * 
from stats.frequency import BLOCK_SIZE
from stats.distributions import * 

from scipy.stats import chi2 as s_chi2, norm as s_norm, kstest as s_kstest

TABLE = r'''
\begin{table}[H]\caption[%(title)s]{%(desc)s}
\begin{center}
\begin{tabular}{%(alignment)s}
%(rows)s
\end{tabular}
\end{center}
\end{table}
'''

HEADER_O_STR = r'Method | Min | Exp. | Max'
HEADER_O = tuple(map(lambda t: t.strip(), HEADER_O_STR.split('|')))

HEADER_STR = r'Prefix | BS | $f$ | $\chi^2$ | $\nabla^2\chi^2$ | KS'
HEADER = tuple(map(lambda t: t.strip(), HEADER_STR.split('|')))


def build_table(title, desc, headers, rows, align=None):
    assert rows   , "No rows!"
    assert len(headers) == len(rows[0]), "len: headers != columns"
    headers = tuple(headers)
    rows = tuple(rows)
    TABLE_LINE = ' & '.join(r'%s' for c in headers) + r' \\ \hline'
    return TABLE % {'title': title, 'desc': desc, 
                    'cols': len(rows[0]), 
                    'alignment': align or '|'.join('c' for col in rows[0]),
                    'rows' : '\n'.join(TABLE_LINE % col for col in ((headers,)+rows))}

def tablelize_float(f, _min=-1*sys.maxint, _max=sys.maxint):
    _str = '%.3f' % f
    if (f > _max):
        _str = r'\mathbf{\blacktriangle\ %s\ \blacktriangle}' % _str
    if (f < _min):
        _str = r'\mathbf{\blacktriangledown\ %s\ \blacktriangledown}' % _str
    return '$%s$' % _str



polys  = Polynomial().query(
    'SELECT * FROM polynomial WHERE name LIKE %s ORDER BY id', ('%',))


for poly in polys:
    print '!', poly.name

    for base in (2, 3, 5, 7, 10):
        print '> base:', base
        if os.path.isfile('tables/%s-%i.tex' % (poly.name, base)):
            print '! already exists'
            continue
            

        groups = FrequencyGroup().query(
            'SELECT * FROM frequencygroup WHERE polynomial_id=%s AND base=%s' + \
                ' ORDER BY runlength DESC LIMIT 3', 
            (poly.id,base))
        if not groups:
            print '! skipping'
            continue

        freqs = {}
        for group in groups:
            print '> runlength:', group.runlength

            freqs[group.runlength] = [] #[(0, {0: BLOCK_SIZE - group.runlength + 1})]
            blocks = FrequencyBlock().query(
                'SELECT * FROM frequencyblock WHERE frequency_group_id=%s AND ' + \
                    '(block_id = 0 OR (block_id+1) %% 16 = 0) ' + \
                    'ORDER BY block_id', 
                (group.id,))

            print '>> ordering'

            ordered = {}
            prev = 0
            lst  = []
            b = None
            for b in blocks:
                if b.block_id != prev:
                    print '>> %6i %10i' % (b.block_id, len(lst))
                    ordered[prev] = list(lst)
                    lst = []
                    prev = b.block_id
                lst.append(b)

            for bid in sorted(ordered):
                digits = ordered[bid]
                freqs[group.runlength] += [(bid, dict((d.digit, d.count) 
                                                      for d in digits))]

        if not freqs.keys():
            print '! skipping - no keys'
            continue

        rows = []
        rl = max(freqs.keys())
        print poly.name, poly.id, tuple(len(freqs[_rl]) for _rl in freqs)

        F_min = F_exp = F_max = None
        KS_min = KS_exp = KS_max = None
        chi2_min = chi2_exp = chi2_max = None
        psi_min = psi_exp = psi_max = None

        ks_list = []
        for i in xrange(len(freqs[rl])):

            # block id, prefix length and sample size
            bid, freq = freqs[rl][i]
            prefix    = bid * BLOCK_SIZE
            blockn    = prefix + BLOCK_SIZE
            outcomes  = base**rl

            ''' simple frequencies '''
            F, (F_min, F_exp, F_max) = frequencies(freq, blockn, outcomes)
            F_str = tablelize_float(F, _min=F_min, _max=F_max)

            ''' Kolminov Smirnov '''
            KS, (KS_min, KS_exp, KS_max) = kolmogorov(freq, blockn, outcomes)
            KS_str     = tablelize_float(KS, _min=KS_min, _max=KS_max)
            
            ''' Chi^2 '''
            # format: freqs[runlength][block_id] == (block_id, block_freqs)
            chi2s    = [ chi_square(freqs[_rl][i][1], blockn - _rl + 1, base**_rl) for 
                         _rl in xrange(rl-2, rl+1) ]
            chi2, (chi2_min,chi2_exp,chi2_max) = chi2s[2]
            chi2_str     = tablelize_float(chi2, _min=chi2_min, _max=chi2_max)

            ''' Nable Chi^2 '''
            # compute nable chi square (here called psi)
            psi, (psi_min,psi_exp,psi_max) = psi_value((chi2s[0][0], chi2s[1][0], chi2s[2][0]), 
                                                       blockn, outcomes, rl, base)
            psi_str = tablelize_float(psi, _min=psi_min, _max=psi_max)

            # generate table row
            row = (r'$2^{%.2f}$' % (21+log2(bid+1)), 
                   '$%i$' % rl, 
                   F_str, 
                   chi2_str, #chi2_exp_str, 
                   psi_str,  #psi_exp_str,
                   KS_str) #KS_exp_str)
            rows += [row]

        # build string
        F_min_str = tablelize_float(F_min)
        F_exp_str = tablelize_float(F_exp)
        F_max_str = tablelize_float(F_max)

        chi2_min_str = tablelize_float(chi2_min)
        chi2_exp_str = tablelize_float(chi2_exp)
        chi2_max_str = tablelize_float(chi2_max)

        psi_min_str = tablelize_float(psi_min)
        psi_exp_str = tablelize_float(psi_exp)
        psi_max_str = tablelize_float(psi_max)

        KS_min_str = tablelize_float(KS_min)
        KS_exp_str = tablelize_float(KS_exp)
        KS_max_str = tablelize_float(KS_max)

        # build and write table
        rows_o  = (
            (r'$f$',              F_min_str,    F_exp_str,    F_max_str),
            (r'$\chi^2$',         chi2_min_str, chi2_exp_str, chi2_max_str), 
            (r'$\nabla^2\chi^2$', psi_min_str,  psi_exp_str,  psi_max_str), 
            ('KS',                KS_min_str,   KS_exp_str,   KS_max_str)
            )
        title = '%s, Base %i - Confidence Intervals' % (poly.name.title(), base)
        desc  = r'%s \\ (%s)' % (title, r'\%s}' % poly.name.replace('-','{'))
        table_o = build_table(title, desc, HEADER_O, rows_o, align='l|r|r|r')

        title = '%s, Base %i - Results' % (poly.name.title(), base)
        desc  = r'%s \\ (%s)' % (title, r'\%s}' % poly.name.replace('-','{'))
        table = build_table(title, desc, HEADER, rows)
        #print table_o
        #print table

        file('tables/%s-%i.tex' % (poly.name, base) ,'w').write(table_o + '\n' + table)


