import os, time, gmpy

from threading import Timer
from aws import s3
from database import * 
from roots import lm
from stats.frequency import absolute_freqs_n_ng, is_uniform, digits_mpf, BLOCK_SIZE


BASES         = (2,3,5,7,10)
PRECISION     = 19 # 2*3**18 ~= 2**(29.5)
RUNLENGTH_MAX = 10


def get_worker_id():
    return '%s-%i' % (os.uname()[1], os.getpid())

def new_process(f):
    pid = os.fork()
    if pid:
        return
    f()

def process_job_root(job, poly):
    job.message = 'calculating root'
    job.save()
    N, root = lm.laguerres_method(poly.polynomial, PRECISION)
    # todo: upload root to amazon
    return root


def process_job_freqs_base(job, poly, base):
    print '='*12, poly.name, 'BASE', base, '='*12
    job.progress = base / float(max(BASES))
    job.message = 'freqs-base:%i' % base
    job.save()

    freqs = absolute_freqs_n_ng(s3.get('%s-b%i' % (poly.name, base)), base)

    for runlength,blocks in freqs:
        print '-'*12, runlength, '-'*12

        freq_group = FrequencyGroup(polynomial_id=poly.id, 
                                    base=base,
                                    runlength=runlength)
        freq_group.save()

        path = '/var/tmp/roots/db/%s-%s-%s.sql' % (poly.name, base, runlength)
        fd = file(path + '.part','w')

        block_id,block = None, None
        for block_id,block in blocks:
            for digit in block.keys():
                FrequencyBlock(frequency_group_id=freq_group.id, 
                               block_id=block_id,
                               digit=digit, count=long(block[digit])).save_to_file(fd)
        fd.close()
        os.rename(path + '.part', path)

        N = block_id*(BLOCK_SIZE+1)
        if not is_uniform(block, N, base, runlength) or runlength >= RUNLENGTH_MAX:
            break


def process_base(job, poly, root, base):
    job.progress = base / float(max(BASES))
    job.message  = 'b%i' % base
    job.save()
    digits = digits_mpf(root, base)
    s3.put('%s-b%i' % (poly.name, base), digits)


def process_bases(job, poly, root):
    map(lambda base: process_base(job, poly, root, base), BASES)


def process_poly_base():
    db_commit(True)

    # try to fetch polynomial with no root
    job = Job().query_one('SELECT * FROM job WHERE type=%s AND status=%s ORDER BY random() LIMIT 1', ('base', 'waiting',))
    if not job:
        return

    # configure job with initial info
    job.status = 'processing'
    job.save()
    job.worker = get_worker_id()
    job.time_start = TimestampFromTicks(time.time())
    job.progress = 0.0
    job.message = 'initializing'
    job.save()

    # check if we got it (this should actually be in a transaction, but its not)
    job = Job().query_one('SELECT * FROM job WHERE id=%s AND status=%s AND worker=%s LIMIT 1', 
                          (job.id, 'processing', get_worker_id()))
    if not job:
        return None

    # fetch polynomial
    poly = Polynomial().query_one('SELECT * FROM polynomial WHERE id=%s', (job.polynomial_id,))

    # calculate root
    job.message = 'downloading root..'
    job.save()
    root = s3.get_mpf(poly.name)
    
    process_bases(job, poly, root)

    job.time_stop = TimestampFromTicks(time.time())
    job.status  = 'done'
    job.message = 'done'
    job.progress = 1.0
    job.save()

    return poly


def process_root():
    poly = Polynomial().query_one('SELECT * FROM polynomial WHERE has_root=%s ORDER BY id LIMIT 1', (False,))
    if not poly:
        return False

    print poly.name, poly.polynomial

    poly.has_root = True
    poly.save()

    print 'computing..'
    N, root = lm.laguerres_method(poly.polynomial, PRECISION)
    print 'uploading..'
    s3.put_mpf(poly.name, root)
    
    poly.precision = N
    poly.save()

    return True
    


def process_stats():
    # try to fetch polynomial with no root
    job = Job().query_one('SELECT * FROM job WHERE type=%s AND status=%s ORDER BY random() LIMIT 1', ('stats', 'waiting',))
    if not job:
        return

    # configure job with initial info
    job.status = 'processing'
    job.save()
    job.worker = get_worker_id()
    job.time_start = TimestampFromTicks(time.time())
    job.progress = 0.0
    job.message = 'initializing'
    job.save()

    # check if we got it (this should actually be in a transaction, but its not)
    job = Job().query_one('SELECT * FROM job WHERE id=%s AND status=%s AND worker=%s LIMIT 1', 
                          (job.id, 'processing', get_worker_id()))
    if not job:
        return None

    # fetch polynomial
    poly = Polynomial().query_one('SELECT * FROM polynomial WHERE id=%s', (job.polynomial_id,))

    # calculate freqs
    map(lambda b : process_job_freqs_base(job, poly, b), BASES)

    job.status  = 'done'
    job.message = 'done'
    job.progess = 1.0
    job.time_stop = TimestampFromTicks(time.time())
    job.save()

    return job.id
