import gmpy, gc, time
from numpy import matrix, array, dot, linalg, identity
from threading import Timer, activeCount
from roots.mp import mp_solve
from roots import PolyLib

MP_SOLVE_PREC = 15


def read_polynomial(polynomial):
    ''' read polynomial from string format '''
    # extract coefficients                                                              
    poly  = PolyLib.kill_multiple_roots(PolyLib.from_string(polynomial))
    coefs = PolyLib.get_coefs(poly)
    N = poly.degree
    # compute replacement matrix                                                        
    R = matrix([[ gmpy.mpz(
        int(r == 0)*-1*int(coefs[N-c-1]) +    # -1 * coef in first row                  
        int(r == c or r-1 == c)                # +1 in diagonal and one row under       
        ) for c in xrange(N) ] for r in xrange(N) ])
    return N, R


def picarte_inv(d, itr):
    ''' compute picarte inverse '''
    i, ii   = gmpy.mpz(1), gmpy.mpz(2)
    r, r2r  = gmpy.mpz(2), gmpy.mpz(0)
    inv     = gmpy.mpz(0)
    for _itr in xrange(itr):
        # square now so r is free
        r2r  = r ** 2
        # overview r
        r   += ii
        # 2^i x_i 
        inv *= r
        # floor ( r_i^2 / b )
        qou = r2r / d
        r   = r2r-qou*d
        # add quotient to inverse
        inv += qou
        # update i and ii
        i  <<= 1
        ii **= 2
        gc.collect()
    return inv
        

def picarte_div(n, d, prec):
    ''' multiplication using picarte inverse '''
    n = n * picarte_inv(d, prec)
    noise = gmpy.mpz(2)**(prec-2)
    retval = n >> (noise*3)
    del noise
    gc.collect()
    return retval



CUTOFF  = gmpy.mpz(2)**20
THREADS = 5
def matrix_square(mat):
    def run(r, c):
        new_mat[r,c] = sum( mat[r,k] * mat[k,c] for k in xrange(N) )
    threads = []

    if mat[0,0].numdigits(2) < CUTOFF:
        compute = run
    else:
        def compute(r, c):
            thread = Timer(0, run, [r,c])
            threads.append(thread)
            while activeCount() >= THREADS:
                time.sleep(1)
            thread.start()            

    N, _ = mat.shape
    new_mat = matrix( [ [ gmpy.mpz(0) for col in xrange(N) ] for row in xrange(N) ] )
    for r in xrange(N):
        if r > 20:
            print r
        for c in xrange(N):
            compute(r,c)

    map(lambda t: t.join(), threads)
    return new_mat

def matrix_power(mat, N):
    for i in xrange(N):
        if i > 20:
            print i
        mat = matrix_square(mat)
    return mat


def power_method(polynomial, prec, use_mp=False):
    ''' compute root using gupta-mittal '''
    assert prec > 0

    print '> matrix'


    N, R = read_polynomial(polynomial)

    #A0 = matrix_power(R, prec) # use this for multithreaded power
    A0 = linalg.matrix_power(R, 2**prec)
    A1 = dot(A0, R)
    
    A0_00 = A0[0,0]
    A0_10 = A0[1,0]
    A1_00 = A1[0,0]
    A1_10 = A1[1,0]

    # collect garbage
    del N, R, A0, A1
    gc.collect()

    print '> division'

    # compute digits and reference digits
    b0 = picarte_div(A0_00, A0_10, prec+2)
    b1 = picarte_div(A1_00, A1_10, prec+2)

    # compare bits (optimise?)
    b0 = b0.digits(2)
    b1 = b1.digits(2)
    gc.collect()

    endpos = 0
    for i in xrange(len(b0)):
        if not b0[i] == b1[i]:
            break
        endpos = i

    # make sure we got it right!
    assert b0[:endpos] == b1[:endpos]
    del b1
    
    
    # make sure we got the same as mp_solve (and use their floating point)
    mp_prec = min(prec, MP_SOLVE_PREC)
    _, mp_root = mp_solve(polynomial, mp_prec)
    mp_data = mp_root.digits(2)
    dec_pos = mp_data.index('.')

    root_data = b0[:dec_pos] + '.' + b0[dec_pos:endpos]

    checkpos = min(endpos, 2**mp_prec) - 16 # ignore unstabalised mp solve part

    assert root_data[:checkpos] == mp_data[:checkpos]

    return (endpos, gmpy.mpf(root_data, endpos+1, 2))



