import os, time, gmpy, sys, numpy, threading

from timings import time_func

from warnings import warn
from c_modules.freqs import _freqs


BLOCK_SIZE     = 2**21
UNIFORM_CUTOFF = 0.05


''' compute freqs of reverse data efficiently '''
def reversed_freqs(data, freqs, base, runlength):
    def get_digits(i):
        return digits(i, base, runlength)
    N, total = freqs[-1]
    N = len(freqs)
    #print N, total
    rev_freqs = [ None for i in xrange(N) ]
    rev_freqs[N-1] = (N-1,dict((int(gmpy.mpz(get_digits(k)[::-1],base)),total[k]) for k in total))
    for i,block in reversed(freqs[:-1]):
        rev_block = {}
        for d in block.keys():
            rev_block[get_digits(d)[::-1]] = total[d] - block[d]
        block_end = (i+1)*BLOCK_SIZE
        for j in xrange(block_end-runlength+1, block_end):
            rev_block[data[j:j+runlength][::-1]] -= 1
        rev_freqs[N-i-2] = (N-i-2, 
                            dict((int(gmpy.mpz(k,base)),v) 
                                 for k,v in rev_block.items()))
    return rev_freqs

def test_reversed_freqs():
    root = gmpy.mpf(gmpy.pi(BLOCK_SIZE), 1024**2*32)**64
    for base in (2,3,5,7,10):
        path = 'test_reversed_freqs-b%i' % base
        #if os.path.isfile(path):
        #data = file(path).read()
        #else:
        data = digits_mpf(root, base)
        #file(path,'w').write(data)
        data = data[:len(data) - len(data) % BLOCK_SIZE]

        func0 = time_func(lambda *args : list(absolute_freqs_rsn(*args)))
        func1 = time_func(lambda *args : list(reversed_freqs(*args)))
        for runlength in (1,2,3,4):
            print 'b%2i / r%2i, blocks: %3i' % (base, runlength, len(data)/BLOCK_SIZE)
            t0, freqs      = func0(data, base, runlength)
            #t0, rev_freqs0 = func0(data[::-1], base, runlength)
            t1, rev_freqs1 = func1(data, freqs, base, runlength)
            t2, rev_freqs2 = func1(data[::-1], rev_freqs1, base, runlength)
            print 'old: %8.4f' % t0
            print 'new: %8.4f' % t1
            assert list(freqs) == list(rev_freqs2)


''' compute freqs of lower runlength efficiently '''
def lower_runlength(data, freqs, runlength):
    from_runlength = len(freqs[0].keys()[0])
    assert runlength < from_runlength

    print from_runlength, '->', len(freqs[-1])
    
    blocks = []
    for i in xrange(len(freqs)):
        block = freqs[i]
        # compute run freqs 
        new_block = {}
        for run in block:
            short_run = run[:runlength]
            new_block[short_run] = new_block.get(short_run, 0) + block[run]
        #print '>>', run
        # compensate for last run
        run_start = (i+1) * BLOCK_SIZE - from_runlength
        #print data[run_start: (i+1)*BLOCK_SIZE]
        for i in xrange(1, from_runlength-runlength+1):
            #print '+',data[run_start+i:run_start+i+runlength]
            run = data[run_start+i:run_start+i+runlength]
            new_block[run] = new_block.get(run, 0) + 1
        blocks.append(new_block)
    return blocks


''' convert int to padded str of digits (skip 0 in 08 and 0x in 0x16) '''
def digits(i, base, runlength=None):
    digits = gmpy.mpz(i).digits(base)
    skip = base == 8 and 1 or base == 16 and 2 or 0
    if runlength:
        return digits[skip:].rjust(runlength, '0')
    return digits[skip:]

def digits_mpf(i, base):
    digits = gmpy.mpf(i).digits(base)
    return digits[digits.index('.')+1:digits.rindex('e')]


''' test if freqs are uniformly distributed '''
def is_uniform(freqs, N, base, runlength):
    num_runs = N - runlength + 1
    expected = num_runs / float(base**runlength)
    is_normal = lambda v: (abs(v - expected) > UNIFORM_CUTOFF * expected)
    # check if max is outside range
    if is_normal(freqs and max(freqs.values()) or 0):
        return False
    # check if min is outside range
    if is_normal(len(freqs) == base**runlength and min(freqs.values()) or 0):
        return False
    return True


''' compute absolute frequencies in specific runlength '''
def absolute_freqs_r(data, base, runlengths):
    if not len(data) % BLOCK_SIZE == 0:
        warn('data will be truncated with: %i digits' % (len(data) % BLOCK_SIZE))

    print runlengths

    # prefill freqss with empty lists
    freqss = dict((runlength, []) for runlength in runlengths)
    freqs = dict( (rl, {}) for rl in runlengths )

    prev_t = time.time()
    for i in xrange(len(data)):
        if time.time() - prev_t > 1:
            print max(runlengths), '>%',i / float(len(data))
            prev_t = time.time()
        # add runs to freqs
        for runlength in runlengths:
            # try catch is faster since its usually ok
            freqs[runlength][data[i:i+runlength]] = freqs[runlength].get(data[i:i+runlength], 0) + 1
            # is this the last run of this block?
            if (i+runlength) % BLOCK_SIZE == 0:
                freqss[runlength].append(dict(freqs[runlength]))
    return freqss

''' absolute freqs for one runlength '''
def absolute_freqs_rs(data, base, runlength):
    if not len(data) % BLOCK_SIZE == 0:
        warn('data will be truncated with: %i digits' % (len(data) % BLOCK_SIZE))

    # prefill freqss with empty lists
    blocks = []
    freqs = {}

    prev_t = time.time()
    for i in xrange(len(data)):
        if time.time() - prev_t > 1:
            print '>%',i / float(len(data))
            prev_t = time.time()
        freqs[data[i:i+runlength]] = freqs.get(data[i:i+runlength], 0) + 1
        # is this the last run of this block?
        if (i+runlength) % BLOCK_SIZE == 0:
            blocks.append(dict(freqs))
    return blocks



''' absolute freqs for one runlength - numpy based '''
def absolute_freqs_rsn(data, base, runlength):
    if not len(data) % BLOCK_SIZE == 0:
        warn('data will be truncated with: %i digits' % (len(data) % BLOCK_SIZE))
    assert len(data) >= 2*BLOCK_SIZE, "need at least 2 blocks"

    # first runs are special cased
    runs  = tuple(gmpy.mpz(data[i:i+runlength], base)
                  for i in xrange(0, BLOCK_SIZE-runlength+1))

    # precompute blocks
    block_pos = range(1, len(data)/BLOCK_SIZE)

    # skip last block if incomplete
    if block_pos[-1] + BLOCK_SIZE > len(data):
        block_pos = block_pos[:-1]

    # map function
    def f_map(blocki):
        block = blocki * BLOCK_SIZE

        if (blocki & 15 == 0):
            print blocki, '/', len(block_pos)

        runs = tuple(gmpy.mpz(data[i:i+runlength], base)
                     for i in xrange(block-runlength+1, 
                                     block+BLOCK_SIZE-runlength+1))

        freqs = numpy.bincount(runs)
        return dict((int(digits(d, 10)), freqs[d]) for d in xrange(len(freqs)))


    # compute first block as special case
    runs = tuple(gmpy.mpz(data[i:i+runlength], base)
                 for i in xrange(BLOCK_SIZE-runlength+1))    
    
    freqs = numpy.bincount(runs)
    prev  = dict((int(digits(d, 10)), freqs[d]) for d in xrange(len(freqs)))
    yield (0, dict(prev))

    # apply map function to each block
    for pos in block_pos:
        curr = f_map(pos)
        
        # accumulate
        for k in prev.keys():
            curr[k] = curr.get(k,0) + prev[k]

        yield (pos, dict(curr))
        prev = curr


''' absolute freqs for one runlength - numpy based '''
def absolute_freqs_rsn_ng(data, base, runlength):
    if not len(data) % BLOCK_SIZE == 0:
        warn('data will be truncated with: %i digits' % (len(data) % BLOCK_SIZE))
    assert len(data) >= 2*BLOCK_SIZE, "need at least 2 blocks"

    # first runs are special cased
    #runs  = tuple(gmpy.mpz(data[i:i+runlength], base)
    #              for i in xrange(0, BLOCK_SIZE-runlength+1))

    # precompute blocks
    block_pos = range(1, len(data)/BLOCK_SIZE)

    # skip last block if incomplete
    if block_pos[-1] + BLOCK_SIZE > len(data):
        block_pos = block_pos[:-1]

    # map function
    def f_map(blocki):
        block = blocki * BLOCK_SIZE

        if (blocki & 15 == 0):
            print blocki, '/', len(block_pos)

        print '> runs'
        runs = []
        append, mpz = runs.append, gmpy.mpz
        for i in xrange(block-runlength+1, block+BLOCK_SIZE-runlength+1):
            d = data[i:i+runlength]
            append(mpz(d, base))
        print '< runs'

        print '> freqs'
        freqs = numpy.bincount(runs)
        print len(freqs)
        print '< freqs' 

        return dict((int(digits(d, 10)), freqs[d]) for d in xrange(len(freqs)))


    prev = _freqs.count(data, 0, runlength, base, BLOCK_SIZE) 

    # compute first block as special case
    #runs = tuple(gmpy.mpz(data[i:i+runlength], base)
    #             for i in xrange(BLOCK_SIZE-runlength+1))    
    
    #freqs = numpy.bincount(runs)
    #prev  = dict((int(digits(d, 10)), freqs[d]) for d in xrange(len(freqs)))
    yield (0, dict(prev))

    # apply map function to each block
    for pos in block_pos:
        curr = _freqs.count(data, pos, runlength, base, BLOCK_SIZE)
        #curr = f_map(pos)
        
        # accumulate
        for k in prev.keys():
            curr[k] = curr.get(k,0) + prev[k]

        yield (pos, dict(curr))
        prev = curr


''' compute absolute frequencies - numpy based '''
def absolute_freqs_n_ng(data, base):
    i = 1
    while True:
        yield (i, absolute_freqs_rsn_ng(data, base, i))
        i += 1


''' compute absolute frequencies - numpy based '''
def absolute_freqs_n(data, base):
    i = 1
    while True:
        yield (i, absolute_freqs_rsn(data, base, i))
        i += 1



''' compute absolute frequencies '''
def absolute_freqs(data, base):
    block = 8
    length, runlength, freqs = len(data), 0, {}
    while (not freqs) or is_uniform(freqs[runlength][-1], length, base, runlength):
        freq = absolute_freqs_r(data, base, range(runlength+1, runlength+block+1))
        freqs.update(zip(freq.keys(), freq.values()))
        runlength += block
        block = max(block/2, 4)
    for r in reversed(list(freqs.keys())):
        if not is_uniform(freqs[r][-1], length, base, r):
            del freqs[r]
    return freqs



''' much faster version of absolute_freqs '''
def absolute_freqs_fast(data, base):
    #if base != 2:
    #    return absolute_freqs(data, base)

    length, runlength, freqs = len(data), 8, {}
    while (not freqs) or is_uniform(freqs[-1], length, base, runlength):
        runlength += 4
        freqs = absolute_freqs_rs(data, base, runlength)

    freqss = { runlength: freqs }
    for i in xrange(runlength-1, 0, -1):
        freqss[i] = lower_runlength(data, freqss[i+1], i)

    for r in reversed(list(freqss.keys())):
        if not is_uniform(freqss[r][-1], length, base, r):
            print 'not uniform:', r
            del freqss[r]
    return freqss


''' calculate absolute freqs from mpf '''
def freqs_of_root(root, base):
    data = root.digits(base)
    # 1.1415e1 -> 1415
    dot_pos = data.index('.')
    exp_pos = data.rindex('e')
    # compute freqs
    return absolute_freqs_n(data[dot_pos+1:exp_pos], base)
